#include <assert.h>
#include <chck/overflow/overflow.h>
#include "internal.h"
#include "macros.h"
#include "visibility.h"
#include "compositor.h"
#include "output.h"
#include "view.h"
#include "session/fd.h"
#include "resources/resources.h"
#include "resources/types/region.h"
#include "resources/types/surface.h"

static void
wl_cb_subsurface_set_position(struct wl_client *client, struct wl_resource *resource, int32_t x, int32_t y)
{
   (void)client;
   struct wlc_surface *surface;
   if (!(surface = convert_from_wlc_resource((wlc_resource)wl_resource_get_user_data(resource), "surface")))
      return;

   surface->pending.subsurface_position = (struct wlc_point){x, y};
}

static void
restack_subsurface_relative_to(wlc_resource surface, wlc_resource sibling, int32_t offset)
{

   size_t surface_idx = (size_t)~0, target_idx = (size_t)~0;

   struct wlc_surface *surface_ptr;
   if (!(surface_ptr = convert_from_wlc_resource(surface, "surface")))
      return;

   struct wlc_surface *parent = convert_from_wlc_resource(surface_ptr->parent, "surface");
   wlc_resource *sub;
   chck_iter_pool_for_each(&parent->subsurface_list, sub) {
      if (*sub == surface)
         surface_idx = _I - 1;

      if (*sub == sibling)
         target_idx = _I - 1;
   }

   if (surface_idx == (size_t)~0 || target_idx == (size_t)~0)
      return;

   if (surface_idx < target_idx)
      --target_idx;

   chck_iter_pool_remove(&parent->subsurface_list, surface_idx);
   chck_iter_pool_insert(&parent->subsurface_list, target_idx + offset, &surface);
}

static void
wl_cb_subsurface_place_above(struct wl_client *client, struct wl_resource *resource, struct wl_resource *sibling_resource)
{
   (void)client, (void)resource, (void)sibling_resource;

   wlc_resource surface_res = (wlc_resource)wl_resource_get_user_data(resource);
   wlc_resource sibling_res = (wlc_resource)wl_resource_get_user_data(sibling_resource);

   restack_subsurface_relative_to(surface_res, sibling_res, 0);
}

static void
wl_cb_subsurface_place_below(struct wl_client *client, struct wl_resource *resource, struct wl_resource *sibling_resource)
{
   (void)client, (void)resource, (void)sibling_resource;

   wlc_resource surface_res = (wlc_resource)wl_resource_get_user_data(resource);
   wlc_resource sibling_res = (wlc_resource)wl_resource_get_user_data(sibling_resource);

   restack_subsurface_relative_to(surface_res, sibling_res, 1);
}

static void
recursive_set_subsurface_parent_sync_state(struct wlc_surface *surface, bool state)
{
   if (!surface || surface->synchronized)
      return;

   surface->parent_synchronized = state;

   wlc_resource *r;
   chck_iter_pool_for_each(&surface->subsurface_list, r)
      recursive_set_subsurface_parent_sync_state(convert_from_wlc_resource(*r, "surface"), state);
}

static void
wl_cb_subsurface_set_sync(struct wl_client *client, struct wl_resource *resource)
{
   (void)client;

   struct wlc_surface *surface;
   if (!(surface = convert_from_wlc_resource((wlc_resource)wl_resource_get_user_data(resource), "surface")))
      return;

   if (surface) {
      surface->synchronized = true;
      recursive_set_subsurface_parent_sync_state(surface, true);
   }
}

static void
wl_cb_subsurface_set_desync(struct wl_client *client, struct wl_resource *resource)
{
   (void)client;

   struct wlc_surface *surface;
   if (!(surface = convert_from_wlc_resource((wlc_resource)wl_resource_get_user_data(resource), "surface")))
      return;

   if (surface) {
      surface->synchronized = false;
      recursive_set_subsurface_parent_sync_state(surface, false);

      struct wlc_surface *parent = convert_from_wlc_resource(surface->parent, "surface");

      if (parent && !parent->synchronized && !parent->parent_synchronized)
         wlc_surface_commit(surface);
   }
}

static const struct wl_subsurface_interface wl_subsurface_implementation = {
   .destroy = wlc_cb_resource_destructor,
   .set_position = wl_cb_subsurface_set_position,
   .place_above = wl_cb_subsurface_place_above,
   .place_below = wl_cb_subsurface_place_below,
   .set_sync = wl_cb_subsurface_set_sync,
   .set_desync = wl_cb_subsurface_set_desync
};

static void
wl_cb_subcompositor_get_subsurface(struct wl_client *client, struct wl_resource *resource, uint32_t id, struct wl_resource *surface_resource, struct wl_resource *parent_resource)
{
   struct wlc_compositor *compositor;
   if (!(compositor = wl_resource_get_user_data(resource)))
      return;

   wlc_resource surface = wlc_resource_from_wl_resource(surface_resource);
   wlc_resource parent = wlc_resource_from_wl_resource(parent_resource);

   if (surface == parent) {
      wl_resource_post_error(resource, WL_SUBCOMPOSITOR_ERROR_BAD_SURFACE, "wl_surface@%d cannot be its own parent", wl_resource_get_id(surface_resource));
      return;
   }

   wlc_resource r;
   if (!(r = wlc_resource_create(&compositor->subsurfaces, client, &wl_subsurface_interface, wl_resource_get_version(resource), 1, id)))
      return;

   wlc_resource_implement(r, &wl_subsurface_implementation, (void*)surface);
   wlc_surface_set_parent(convert_from_wlc_resource(surface, "surface"), convert_from_wlc_resource(parent, "surface"));
}

static void
wl_cb_subcompositor_destroy(struct wl_client *client, struct wl_resource *resource)
{
   (void)client;
   wl_resource_destroy(resource);
}

static const struct wl_subcompositor_interface wl_subcompositor_implementation = {
   .destroy = wl_cb_subcompositor_destroy,
   .get_subsurface = wl_cb_subcompositor_get_subsurface
};

static void
wl_subcompositor_bind(struct wl_client *client, void *data, uint32_t version, uint32_t id)
{
   struct wl_resource *resource;
   if (!(resource = wl_resource_create_checked(client, &wl_subcompositor_interface, version, 1, id)))
      return;

   wl_resource_set_implementation(resource, &wl_subcompositor_implementation, data, NULL);
}

static void
wl_cb_surface_create(struct wl_client *client, struct wl_resource *resource, uint32_t id)
{
   struct wlc_compositor *compositor;
   if (!(compositor = wl_resource_get_user_data(resource)))
      return;

   wlc_resource r;
   if (!(r = wlc_resource_create(&compositor->surfaces, client, &wl_surface_interface, wl_resource_get_version(resource), 3, id)))
      return;

   wlc_resource_implement(r, wlc_surface_implementation(), compositor);

   struct wlc_surface_event ev = { .surface = convert_from_wlc_resource(r, "surface"), .type = WLC_SURFACE_EVENT_CREATED };
   wl_signal_emit(&wlc_system_signals()->surface, &ev);
}

static void
wl_cb_region_create(struct wl_client *client, struct wl_resource *resource, uint32_t id)
{
   struct wlc_compositor *compositor;
   if (!(compositor = wl_resource_get_user_data(resource)))
      return;

   wlc_resource r;
   if (!(r = wlc_resource_create(&compositor->regions, client, &wl_region_interface, wl_resource_get_version(resource), 3, id)))
      return;

   wlc_resource_implement(r, wlc_region_implementation(), wl_resource_get_user_data(resource));
}

static const struct wl_compositor_interface wl_compositor_implementation = {
   .create_surface = wl_cb_surface_create,
   .create_region = wl_cb_region_create
};

static void
wl_compositor_bind(struct wl_client *client, void *data, uint32_t version, uint32_t id)
{
   struct wl_resource *r;
   if (!(r = wl_resource_create_checked(client, &wl_compositor_interface, version, 3, id)))
      return;

   wl_resource_set_implementation(r, &wl_compositor_implementation, data, NULL);
}

static void
cb_idle_vt_switch(void *data)
{
   struct wlc_compositor *compositor = data;
   wlc_fd_activate_vt(compositor->state.vt);
   compositor->state.vt = 0;
   wl_event_source_remove(compositor->state.idle);
   compositor->state.idle = NULL;
}

static void
activate_tty(struct wlc_compositor *compositor)
{
   if (compositor->state.tty != ACTIVATING)
      return;

   compositor->state.tty = IDLE;
   wlc_fd_activate();
}

static void
deactivate_tty(struct wlc_compositor *compositor)
{
   if (compositor->state.tty != DEACTIVATING)
      return;

   // check that all outputs are surfaceless
   struct wlc_output *o;
   chck_pool_for_each(&compositor->outputs.pool, o) {
      if (o->bsurface.display)
         return;
   }

   compositor->state.tty = IDLE;

   if (compositor->state.vt != 0) {
      compositor->state.idle = wl_event_loop_add_idle(wlc_event_loop(), cb_idle_vt_switch, compositor);
   } else {
      wlc_fd_deactivate();
   }
}

static void
respond_tty_activate(struct wlc_compositor *compositor)
{
   if (compositor->state.tty == ACTIVATING) {
      activate_tty(compositor);
   } else if (compositor->state.tty == DEACTIVATING) {
      deactivate_tty(compositor);
   }
}

static void
activate_event(struct wl_listener *listener, void *data)
{
   struct wlc_compositor *compositor;
   except(compositor = wl_container_of(listener, compositor, listener.activate));

   struct wlc_activate_event *ev = data;
   if (!ev->active) {
      compositor->state.tty = DEACTIVATING;
      compositor->state.vt = ev->vt;
      chck_pool_for_each_call(&compositor->outputs.pool, wlc_output_set_backend_surface, NULL);
      deactivate_tty(compositor);
   } else {
      compositor->state.tty = ACTIVATING;
      compositor->state.vt = 0;
      activate_tty(compositor);
      wlc_backend_update_outputs(&compositor->backend, &compositor->outputs.pool);
      chck_pool_for_each_call(&compositor->outputs.pool, wlc_output_set_sleep_ptr, false);
   }
}

static void
terminate_event(struct wl_listener *listener, void *data)
{
   (void)data;
   struct wlc_compositor *compositor;
   except(compositor = wl_container_of(listener, compositor, listener.terminate));
   wlc_compositor_terminate(compositor);
}

static void
xwayland_event(struct wl_listener *listener, void *data)
{
   bool activated = *(bool*)data;

   struct wlc_compositor *compositor;
   except(compositor = wl_container_of(listener, compositor, listener.xwayland));

   if (activated) {
      wlc_xwm(&compositor->xwm, &compositor->seat);
   } else {
      wlc_xwm_release(&compositor->xwm);
   }

   if (!compositor->state.ready) {
      WLC_INTERFACE_EMIT(compositor.ready);
      compositor->state.ready = true;
   }
}

static void
attach_surface_to_view_or_create(struct wlc_compositor *compositor, struct wlc_surface *surface, enum wlc_surface_role type, wlc_resource role)
{
   assert(compositor && surface && type < WLC_SURFACE_ROLE_LAST);

   struct wlc_view *view;
   if (!(view = wlc_compositor_view_for_surface(compositor, surface)))
      return;

   // views without role are ok.
   if (!role)
      return;

   wlc_resource *res[WLC_SURFACE_ROLE_LAST] = {
      &view->shell_surface,
      &view->xdg_surface,
      &view->xdg_toplevel,
      &view->custom_surface,
   };

   const char *name[WLC_SURFACE_ROLE_LAST] = {
      "shell-surface",
      "xdg-surface",
      "xdg-toplevel",
      "custom-surface",
   };

   *res[type] = role;

   if (type != WLC_CUSTOM_SURFACE)
      wl_resource_set_user_data(wl_resource_from_wlc_resource(role, name[type]), (void*)convert_to_wlc_handle(view));
}

static void
attach_popup_to_view_or_create(struct wlc_compositor *compositor, struct wlc_surface *surface, struct wlc_surface *parent, wlc_resource role)
{
   assert(compositor && surface && parent);

   struct wlc_view *view;
   if (!(view = wlc_compositor_view_for_surface(compositor, surface)))
      return;

   view->xdg_popup = role;
   wlc_view_set_parent_ptr(view, convert_from_wlc_handle(parent->view, "view"));
   wlc_view_set_type_ptr(view, WLC_BIT_POPUP, true);
   wl_resource_set_user_data(wl_resource_from_wlc_resource(role, "xdg-popup"), (void*)convert_to_wlc_handle(view));
}

static void
surface_event(struct wl_listener *listener, void *data)
{
   struct wlc_compositor *compositor;
   except(compositor = wl_container_of(listener, compositor, listener.surface));

   struct wlc_surface_event *ev = data;
   switch (ev->type) {
      case WLC_SURFACE_EVENT_REQUEST_VIEW_ATTACH:
         attach_surface_to_view_or_create(compositor, ev->surface, ev->attach.type, ev->attach.role);
         break;

      case WLC_SURFACE_EVENT_REQUEST_VIEW_POPUP:
         attach_popup_to_view_or_create(compositor, ev->surface, ev->popup.parent, ev->popup.role);
         break;

      case WLC_SURFACE_EVENT_DESTROYED:
      {
         struct wlc_view *v;
         chck_pool_for_each(&compositor->views.pool, v) {
            if (v->parent == ev->surface->view)
               wlc_view_set_parent_ptr(v, NULL);
         }

         struct wlc_surface *s;
         chck_pool_for_each(&compositor->surfaces.pool, s) {
            if (s->parent == convert_to_wlc_resource(ev->surface))
               wlc_surface_set_parent(s, NULL);
         }
      }
      break;

      default: break;
   }
}

static struct wlc_output*
get_surfaceless_output(struct wlc_compositor *compositor)
{
   struct wlc_output *o;
   chck_pool_for_each(&compositor->outputs.pool, o) {
      if (!o->bsurface.display)
         return o;
   }
   return NULL;
}

static void
active_output(struct wlc_compositor *compositor, struct wlc_output *output)
{
   assert(compositor);

   wlc_dlog(WLC_DBG_FOCUS, "focus output %" PRIuWLC " %" PRIuWLC, compositor->active.output, convert_to_wlc_handle(output));

   if (compositor->active.output == convert_to_wlc_handle(output))
      return;

   if (compositor->active.output)
      WLC_INTERFACE_EMIT(output.focus, compositor->active.output, false);

   wlc_output_schedule_repaint(convert_from_wlc_handle(compositor->active.output, "output"));
   compositor->active.output = convert_to_wlc_handle(output);

   if (compositor->active.output) {
      WLC_INTERFACE_EMIT(output.focus, compositor->active.output, true);
      wlc_output_schedule_repaint(output);
   }
}

static void
add_output(struct wlc_compositor *compositor, struct wlc_backend_surface *bsurface, struct wlc_output_information *info)
{
   assert(compositor && bsurface && info);

   struct wlc_output *output;
   if (!(output = get_surfaceless_output(compositor)))
      output = wlc_handle_create(&compositor->outputs);

   if (!output) {
      wlc_backend_surface_release(bsurface);
      return;
   }

   wlc_output_set_information(output, info);

   if (WLC_INTERFACE_EMIT_EXCEPT(output.created, false, convert_to_wlc_handle(output))) {
      wlc_output_terminate(output);
      return;
   }

   output->state.created = true;

   wlc_output_set_backend_surface(output, bsurface);

   if (!compositor->active.output)
      active_output(compositor, output);

   wlc_output_schedule_repaint(output);
   wlc_log(WLC_LOG_INFO, "Added output (%" PRIuWLC ")", convert_to_wlc_handle(output));
}

static void
remove_output(struct wlc_compositor *compositor, struct wlc_output *output)
{
   assert(compositor && output);

   struct wlc_output *o, *alive = NULL;
   chck_pool_for_each(&compositor->outputs.pool, o) {
      if (!o->bsurface.display || o == output)
         continue;

      alive = o;
      break;
   }

   if (compositor->active.output == convert_to_wlc_handle(output)) {
      compositor->active.output = 0; // make sure we don't redraw
      active_output(compositor, alive);
   }

   WLC_INTERFACE_EMIT(output.destroyed, convert_to_wlc_handle(output));
   wlc_output_set_backend_surface(output, NULL);

   wlc_log(WLC_LOG_INFO, "Removed output (%" PRIuWLC ")", convert_to_wlc_handle(output));

   if (compositor->state.terminating && !alive)
      wlc_compositor_terminate(compositor);
}

static void
output_event(struct wl_listener *listener, void *data)
{
   struct wlc_output_event *ev = data;

   struct wlc_compositor *compositor;
   except(compositor = wl_container_of(listener, compositor, listener.output));

   switch (ev->type) {
      case WLC_OUTPUT_EVENT_ADD:
         add_output(compositor, ev->add.bsurface, ev->add.info);
         break;

      case WLC_OUTPUT_EVENT_ACTIVE:
         active_output(compositor, ev->active.output);
         break;

      case WLC_OUTPUT_EVENT_REMOVE:
         remove_output(compositor, ev->remove.output);
         break;

      case WLC_OUTPUT_EVENT_UPDATE:
         wlc_backend_update_outputs(&compositor->backend, &compositor->outputs.pool);
         break;

      case WLC_OUTPUT_EVENT_SURFACE:
         respond_tty_activate(compositor);
         break;
   }
}

static void
focus_event(struct wl_listener *listener, void *data)
{
   struct wlc_compositor *compositor;
   except(compositor = wl_container_of(listener, compositor, listener.focus));

   struct wlc_focus_event *ev = data;
   switch (ev->type) {
      case WLC_FOCUS_EVENT_OUTPUT:
         active_output(compositor, ev->output);
         break;

      default: break;
   }
}

struct wlc_view*
wlc_compositor_view_for_surface(struct wlc_compositor *compositor, struct wlc_surface *surface)
{
   struct wlc_view *view;
   if (!(view = convert_from_wlc_handle(surface->view, "view")) && !(view = wlc_handle_create(&compositor->views)))
      return NULL;

   wlc_surface_attach_to_output(surface, convert_from_wlc_handle(compositor->active.output, "output"), wlc_surface_get_buffer(surface));
   wlc_surface_attach_to_view(surface, view);
   return view;
}

// XXX: We do not currently expose compositor to public API.
//      So we use static variable here for some public api functions.
//
//      Never use this variable anywhere else.
static struct wlc_compositor *_g_compositor;

WLC_API const wlc_handle*
wlc_get_outputs(size_t *out_memb)
{
   assert(_g_compositor);

   if (out_memb)
      *out_memb = 0;

   // Allocate linear array which we then return
   free(_g_compositor->tmp.outputs);
   if (!(_g_compositor->tmp.outputs = chck_malloc_mul_of(_g_compositor->outputs.pool.items.count, sizeof(wlc_handle))))
      return NULL;

   {
      size_t i = 0;
      struct wlc_output *o;
      chck_pool_for_each(&_g_compositor->outputs.pool, o)
         _g_compositor->tmp.outputs[i++] = convert_to_wlc_handle(o);
   }

   if (out_memb)
      *out_memb = _g_compositor->outputs.pool.items.count;

   return _g_compositor->tmp.outputs;
}

WLC_API WLC_PURE wlc_handle
wlc_get_focused_output(void)
{
   assert(_g_compositor);
   return _g_compositor->active.output;
}

WLC_API WLC_PURE struct xkb_state*
wlc_keyboard_get_xkb_state(void)
{
   assert(_g_compositor);
   return _g_compositor->seat.keyboard.state.xkb;
}

WLC_API WLC_PURE struct xkb_keymap*
wlc_keyboard_get_xkb_keymap(void)
{
   assert(_g_compositor);
   return _g_compositor->seat.keymap.keymap;
}

WLC_API const uint32_t*
wlc_keyboard_get_current_keys(size_t *out_memb)
{
   assert(_g_compositor);
   return chck_iter_pool_to_c_array(&_g_compositor->seat.keyboard.keys, out_memb);
}

WLC_API uint32_t
wlc_keyboard_get_keysym_for_key(uint32_t key, const struct wlc_modifiers *modifiers)
{
   assert(_g_compositor);
   return wlc_keyboard_get_keysym_for_key_ptr(&_g_compositor->seat.keyboard, key, modifiers);
}

WLC_API uint32_t
wlc_keyboard_get_utf32_for_key(uint32_t key, const struct wlc_modifiers *modifiers)
{
   assert(_g_compositor);
   return wlc_keyboard_get_utf32_for_key_ptr(&_g_compositor->seat.keyboard, key, modifiers);
}

WLC_API void
wlc_pointer_get_position_v2(double *out_x, double *out_y)
{
   assert(_g_compositor && out_x && out_y);
   *out_x = _g_compositor->seat.pointer.pos.x;
   *out_y = _g_compositor->seat.pointer.pos.y;
}

WLC_API void
wlc_pointer_get_position(struct wlc_point *out_position)
{
   assert(_g_compositor && out_position);
   out_position->x = _g_compositor->seat.pointer.pos.x;
   out_position->y = _g_compositor->seat.pointer.pos.y;
}

WLC_API void
wlc_pointer_set_position_v2(double x, double y)
{
   assert(_g_compositor);
   _g_compositor->seat.pointer.pos.x = x;
   _g_compositor->seat.pointer.pos.y = y;
}

WLC_API void
wlc_pointer_set_position(const struct wlc_point *position)
{
   assert(_g_compositor && position);
   _g_compositor->seat.pointer.pos.x = position->x;
   _g_compositor->seat.pointer.pos.y = position->y;
}

void
wlc_compositor_terminate(struct wlc_compositor *compositor)
{
   if (!compositor || !_g_compositor)
      return;

   if (!compositor->state.terminating) {
      wlc_log(WLC_LOG_INFO, "Terminating compositor...");
      compositor->state.terminating = true;

      WLC_INTERFACE_EMIT(compositor.terminate);

      if (compositor->outputs.pool.items.count > 0) {
         chck_pool_for_each_call(&compositor->outputs.pool, wlc_output_terminate);
         return;
      }
   }

   wlc_log(WLC_LOG_INFO, "Compositor terminated...");
   wl_signal_emit(&wlc_system_signals()->compositor, NULL);
}

bool
wlc_compositor_is_good(struct wlc_compositor *compositor)
{
   assert(compositor);

   struct wlc_output *o;
   chck_pool_for_each(&compositor->outputs.pool, o) {
      if (o->context.context)
         return true;
   }

   // There were no outputs or all outputs failed context creation
   // wlc.c does this check and will return false in init
   // XXX: Skip this check when we request headless mode
   return false;
}

void
wlc_compositor_release(struct wlc_compositor *compositor)
{
   if (!compositor || !_g_compositor)
      return;

   wl_list_remove(&compositor->listener.activate.link);
   wl_list_remove(&compositor->listener.terminate.link);
   wl_list_remove(&compositor->listener.xwayland.link);
   wl_list_remove(&compositor->listener.surface.link);
   wl_list_remove(&compositor->listener.output.link);
   wl_list_remove(&compositor->listener.focus.link);

   wlc_backend_release(&compositor->backend);
   wlc_shell_release(&compositor->shell);
   wlc_xdg_shell_release(&compositor->xdg_shell);
   wlc_custom_shell_release(&compositor->custom_shell);
   wlc_seat_release(&compositor->seat);

   if (compositor->wl.subcompositor)
      wl_global_destroy(compositor->wl.subcompositor);

   if (compositor->wl.compositor)
      wl_global_destroy(compositor->wl.compositor);

   free(_g_compositor->tmp.outputs);
   wlc_source_release(&compositor->outputs);
   wlc_source_release(&compositor->views);
   wlc_source_release(&compositor->surfaces);
   wlc_source_release(&compositor->subsurfaces);
   wlc_source_release(&compositor->regions);

   memset(compositor, 0, sizeof(struct wlc_compositor));
   _g_compositor = NULL;
}

bool
wlc_compositor(struct wlc_compositor *compositor)
{
   assert(wlc_display() && wlc_event_loop() && !_g_compositor);
   memset(compositor, 0, sizeof(struct wlc_compositor));

   if (!wlc_display() || !wlc_event_loop() || _g_compositor) {
      wlc_log(WLC_LOG_ERROR, "wlc_compositor called before wlc_init()");
      abort();
   }

   _g_compositor = compositor;

   compositor->listener.activate.notify = activate_event;
   compositor->listener.terminate.notify = terminate_event;
   compositor->listener.xwayland.notify = xwayland_event;
   compositor->listener.surface.notify = surface_event;
   compositor->listener.output.notify = output_event;
   compositor->listener.focus.notify = focus_event;
   wl_signal_add(&wlc_system_signals()->activate, &compositor->listener.activate);
   wl_signal_add(&wlc_system_signals()->terminate, &compositor->listener.terminate);
   wl_signal_add(&wlc_system_signals()->xwayland, &compositor->listener.xwayland);
   wl_signal_add(&wlc_system_signals()->surface, &compositor->listener.surface);
   wl_signal_add(&wlc_system_signals()->output, &compositor->listener.output);
   wl_signal_add(&wlc_system_signals()->focus, &compositor->listener.focus);

   if (!wlc_source(&compositor->outputs, "output", wlc_output, wlc_output_release, 4, sizeof(struct wlc_output)) ||
       !wlc_source(&compositor->views, "view", wlc_view, wlc_view_release, 32, sizeof(struct wlc_view)) ||
       !wlc_source(&compositor->surfaces, "surface", wlc_surface, wlc_surface_release, 32, sizeof(struct wlc_surface)) ||
       !wlc_source(&compositor->subsurfaces, "subsurface", NULL, NULL, 32, sizeof(struct wlc_resource)) ||
       !wlc_source(&compositor->regions, "region", NULL, wlc_region_release, 32, sizeof(struct wlc_region)))
      goto fail;

   if (!(compositor->wl.compositor = wl_global_create(wlc_display(), &wl_compositor_interface, 3, compositor, wl_compositor_bind)))
      goto compositor_interface_fail;

   if (!(compositor->wl.subcompositor = wl_global_create(wlc_display(), &wl_subcompositor_interface, 1, compositor, wl_subcompositor_bind)))
      goto subcompositor_interface_fail;

   if (!wlc_seat(&compositor->seat) ||
       !wlc_shell(&compositor->shell) ||
       !wlc_xdg_shell(&compositor->xdg_shell) ||
       !wlc_custom_shell(&compositor->custom_shell) ||
       !wlc_backend(&compositor->backend))
      goto fail;

   return true;

compositor_interface_fail:
   wlc_log(WLC_LOG_WARN, "Failed to bind compositor interface");
   goto fail;
subcompositor_interface_fail:
   wlc_log(WLC_LOG_WARN, "Failed to bind subcompositor interface");
   goto fail;
fail:
   wlc_compositor_release(compositor);
   return false;
}
